==> edx-i18n-tools-0.3.3.tar.gz: edx-i18n-tools-0.3.3/i18n/validate.py <==
"""Tests that validate .po files."""

from __future__ import print_function

import codecs
import logging
import os
import struct
import sys
import textwrap

import polib

from i18n.dummy import is_format_message
from i18n.execute import call
from i18n.converter import Converter
from i18n import Runner, config

log = logging.getLogger(__name__)


def validate_po_files(root, report_empty=False):
    """
    Validate all of the po files found in the root directory that are not product of a merge.
    """
    # List of .po files that are the product of a merge (see generate.py).
    merged_files = config.CONFIGURATION.generate_merge.keys()

    for dirpath, __, filenames in os.walk(root):
        for name in filenames:
            __, ext = os.path.splitext(name)
            filename = os.path.join(dirpath, name)

            # Validate only .po files that are not product of a merge (see generate.py).
            # If django-partial.po has a problem, then django.po will also, so don't report it.
            if ext.lower() == '.po' and os.path.basename(filename) not in merged_files:

                # First validate the format of this file
                msgfmt_check_po_file(filename)

                # Check that the translated strings are valid, and optionally
                # check for empty translations. But don't check English.
                if "/locale/en/" not in filename:
                    problems = check_messages(filename, report_empty)
                    report_problems(filename, problems)


def msgfmt_check_po_file(filename):
    """
    Call GNU msgfmt -c on each .po file to validate its format.
    Any errors caught by msgfmt are logged to log.
    """
    # Use relative paths to make output less noisy.
    rfile = os.path.relpath(filename, config.LOCALE_DIR)
    out, err = call('msgfmt -c -o /dev/null {}'.format(rfile), working_directory=config.LOCALE_DIR)
    if err != '':
        log.info('\n' + out)
        log.warning('\n' + err)


def tags_in_string(msg):
    """
    Return the set of tags in a message string.

    Tags includes HTML tags, data placeholders, etc.

    Skips tags that might change due to translations: HTML entities, <abbr>,
    and so on.

    """
    def is_linguistic_tag(tag):
        """Is this tag one that can change with the language?"""
        if tag.startswith("&"):
            return True
        if any(x in tag for x in ["<abbr>", "<abbr ", "</abbr>"]):
            return True
        return False

    __, tags = Converter().detag_string(msg)
    return set(t for t in tags if not is_linguistic_tag(t))


def astral(msg):
    """Does `msg` have characters outside the Basic Multilingual Plane?"""
    # Python2 narrow builds present astral characters as surrogate pairs.
    # By encoding as utf32, and decoding DWORDS, we can get at the real code
    # points.
    utf32 = msg.encode("utf32")[4:]         # [4:] to drop the bom
    code_points = struct.unpack("%dI" % (len(utf32) / 4), utf32)
    return any(cp > 0xFFFF for cp in code_points)


def check_messages(filename, report_empty=False):
    """
    Checks messages in `filename` in various ways:

    * Translations must have the same slots as the English.

    * Messages can't have astral characters in them.

    If `report_empty` is True, will also report empty translation strings.

    Returns the problems, a list of tuples. Each is a description, a msgid, and
    then zero or more translations.

    """
    problems = []
    pomsgs = polib.pofile(filename)
    for msg in pomsgs:
        # Check for characters Javascript can't support.
        # https://code.djangoproject.com/ticket/21725
        if astral(msg.msgstr):
            problems.append(("Non-BMP char", msg.msgid, msg.msgstr))

        if is_format_message(msg):
            # LONG_DATE_FORMAT, etc, have %s etc in them, and that's ok.
            continue

        if msg.msgid_plural:
            # Plurals: two strings in, N strings out.
            source = msg.msgid + " | " + msg.msgid_plural
            translation = " | ".join(v for k, v in sorted(msg.msgstr_plural.items()))
            empty = any(not t.strip() for t in msg.msgstr_plural.values())
        else:
            # Singular: just one string in and one string out.
            source = msg.msgid
            translation = msg.msgstr
            empty = not msg.msgstr.strip()

        if empty:
            if report_empty:
                problems.append(("Empty translation", source))
        else:
            id_tags = tags_in_string(source)
            tx_tags = tags_in_string(translation)

            # Check if tags don't match
            if id_tags != tx_tags:
                id_has = u", ".join(u'"{}"'.format(t) for t in id_tags - tx_tags)
                tx_has = u", ".join(u'"{}"'.format(t) for t in tx_tags - id_tags)
                if id_has and tx_has:
                    diff = u"{} vs {}".format(id_has, tx_has)
                elif id_has:
                    diff = u"{} missing".format(id_has)
                else:
                    diff = u"{} added".format(tx_has)
                problems.append((
                    "Different tags in source and translation",
                    source,
                    translation,
                    diff
                ))

    return problems


def report_problems(filename, problems):
    """
    Report on the problems found in `filename`.

    `problems` is a list of tuples as returned by `check_messages`.

    """
    if problems:
        problem_file = filename.replace(".po", ".prob")
        id_filler = textwrap.TextWrapper(width=79, initial_indent="  msgid: ", subsequent_indent=" " * 9)
        tx_filler = textwrap.TextWrapper(width=79, initial_indent="  -----> ", subsequent_indent=" " * 9)
        with codecs.open(problem_file, "w", encoding="utf8") as prob_file:
            for problem in problems:
                desc, msgid = problem[:2]
                prob_file.write(u"{}\n{}\n".format(desc, id_filler.fill(msgid)))
                for translation in problem[2:]:
                    prob_file.write(u"{}\n".format(tx_filler.fill(translation)))
                prob_file.write(u"\n")

        log.error(" %s problems in %s, details in .prob file", len(problems), filename)
    else:
        dup_filename = filename.replace('.po', '.dup')
        if os.path.exists(dup_filename):
            log.warning(" Duplicates found in %s, details in .dup file", dup_filename)
        log.info(" No problems found in %s", filename)


class Validate(Runner):
    """
    The `validate` command for checking that .po files have no errors.
    """

    def add_args(self):
        self.parser.description = "Automatically finds translation errors in all edx-platform *.po files, "\
            "for all languages, unless one or more language(s) is specified to check."

        self.parser.add_argument(
            '-l', '--language',
            type=str,
            nargs='*',
            help="Specify one or more specific language code(s) to check (eg 'ko_KR')."
        )

        self.parser.add_argument(
            '-e', '--empty',
            action='store_true',
            help="Includes empty translation strings in .prob files."
        )

    def run(self, args):
        """
        Main entry point for script
        """
        if args.verbose:
            log_level = logging.INFO
        else:
            log_level = logging.WARNING
        logging.basicConfig(stream=sys.stdout, level=log_level)

        languages = args.language or []

        if not languages:
            root = config.LOCALE_DIR
            validate_po_files(root, args.empty)
            return

        # languages will be a list of language codes; test each language.
        for language in languages:
            root = config.LOCALE_DIR / language
            # Assert that a directory for this language code exists on the system
            if not root.isdir():
                log.error(" %s is not a valid directory.\nSkipping language '%s'", root, language)
                continue
            # If we found the language code's directory, validate the files.
            validate_po_files(root, args.empty)

main = Validate()  # pylint: disable=invalid-name

if __name__ == '__main__':
    print("Validating languages...")
    main()
    print("Finished validating languages")

==> edx-i18n-tools-0.3.6.tar.gz: edx-i18n-tools-0.3.6/i18n/validate.py <==
"""Tests that validate .po files."""

from __future__ import print_function

import codecs
import logging
import os
import struct
import sys
import textwrap

import polib

from i18n.dummy import is_format_message
from i18n.execute import call
from i18n.converter import Converter
from i18n import Runner

log = logging.getLogger(__name__)


def validate_po_files(configuration, locale_dir, root_dir=None, report_empty=False):
    """
    Validate all of the po files found in the root directory that are not product of a merge.
    """
    # List of .po files that are the product of a merge (see generate.py).
    merged_files = configuration.generate_merge.keys()

    for dirpath, __, filenames in os.walk(root_dir if root_dir else locale_dir):
        for name in filenames:
            __, ext = os.path.splitext(name)
            filename = os.path.join(dirpath, name)

            # Validate only .po files that are not product of a merge (see generate.py).
            # If django-partial.po has a problem, then django.po will also, so don't report it.
            if ext.lower() == '.po' and os.path.basename(filename) not in merged_files:

                # First validate the format of this file
                msgfmt_check_po_file(locale_dir, filename)

                # Check that the translated strings are valid, and optionally
                # check for empty translations. But don't check English.
                if "/locale/en/" not in filename:
                    problems = check_messages(filename, report_empty)
                    report_problems(filename, problems)


def msgfmt_check_po_file(locale_dir, filename):
    """
    Call GNU msgfmt -c on each .po file to validate its format.
    Any errors caught by msgfmt are logged to log.
    """
    # Use relative paths to make output less noisy.
    rfile = os.path.relpath(filename, locale_dir)
    out, err = call('msgfmt -c -o /dev/null {}'.format(rfile), working_directory=locale_dir)
    if err != '':
        log.info('\n' + out)
        log.warning('\n' + err)


def tags_in_string(msg):
    """
    Return the set of tags in a message string.

    Tags includes HTML tags, data placeholders, etc.

    Skips tags that might change due to translations: HTML entities, <abbr>,
    and so on.

    """
    def is_linguistic_tag(tag):
        """Is this tag one that can change with the language?"""
        if tag.startswith("&"):
            return True
        if any(x in tag for x in ["<abbr>", "<abbr ", "</abbr>"]):
            return True
        return False

    __, tags = Converter().detag_string(msg)
    return set(t for t in tags if not is_linguistic_tag(t))


def astral(msg):
    """Does `msg` have characters outside the Basic Multilingual Plane?"""
    # Python2 narrow builds present astral characters as surrogate pairs.
    # By encoding as utf32, and decoding DWORDS, we can get at the real code
    # points.
    utf32 = msg.encode("utf32")[4:]         # [4:] to drop the bom
    code_points = struct.unpack("%dI" % (len(utf32) / 4), utf32)
    return any(cp > 0xFFFF for cp in code_points)


def check_messages(filename, report_empty=False):
    """
    Checks messages in `filename` in various ways:

    * Translations must have the same slots as the English.

    * Messages can't have astral characters in them.

    If `report_empty` is True, will also report empty translation strings.

    Returns the problems, a list of tuples. Each is a description, a msgid, and
    then zero or more translations.

    """
    problems = []
    pomsgs = polib.pofile(filename)
    for msg in pomsgs:
        # Check for characters Javascript can't support.
        # https://code.djangoproject.com/ticket/21725
        if astral(msg.msgstr):
            problems.append(("Non-BMP char", msg.msgid, msg.msgstr))

        if is_format_message(msg):
            # LONG_DATE_FORMAT, etc, have %s etc in them, and that's ok.
            continue

        if msg.msgid_plural:
            # Plurals: two strings in, N strings out.
            source = msg.msgid + " | " + msg.msgid_plural
            translation = " | ".join(v for k, v in sorted(msg.msgstr_plural.items()))
            empty = any(not t.strip() for t in msg.msgstr_plural.values())
        else:
            # Singular: just one string in and one string out.
            source = msg.msgid
            translation = msg.msgstr
            empty = not msg.msgstr.strip()

        if empty:
            if report_empty:
                problems.append(("Empty translation", source))
        else:
            id_tags = tags_in_string(source)
            tx_tags = tags_in_string(translation)

            # Check if tags don't match
            if id_tags != tx_tags:
                id_has = u", ".join(u'"{}"'.format(t) for t in id_tags - tx_tags)
                tx_has = u", ".join(u'"{}"'.format(t) for t in tx_tags - id_tags)
                if id_has and tx_has:
                    diff = u"{} vs {}".format(id_has, tx_has)
                elif id_has:
                    diff = u"{} missing".format(id_has)
                else:
                    diff = u"{} added".format(tx_has)
                problems.append((
                    "Different tags in source and translation",
                    source,
                    translation,
                    diff
                ))

    return problems


def report_problems(filename, problems):
    """
    Report on the problems found in `filename`.

    `problems` is a list of tuples as returned by `check_messages`.

    """
    if problems:
        problem_file = filename.replace(".po", ".prob")
        id_filler = textwrap.TextWrapper(width=79, initial_indent="  msgid: ", subsequent_indent=" " * 9)
        tx_filler = textwrap.TextWrapper(width=79, initial_indent="  -----> ", subsequent_indent=" " * 9)
        with codecs.open(problem_file, "w", encoding="utf8") as prob_file:
            for problem in problems:
                desc, msgid = problem[:2]
                prob_file.write(u"{}\n{}\n".format(desc, id_filler.fill(msgid)))
                for translation in problem[2:]:
                    prob_file.write(u"{}\n".format(tx_filler.fill(translation)))
                prob_file.write(u"\n")

        log.error(" %s problems in %s, details in .prob file", len(problems), filename)
    else:
        dup_filename = filename.replace('.po', '.dup')
        if os.path.exists(dup_filename):
            log.warning(" Duplicates found in %s, details in .dup file", dup_filename)
        log.info(" No problems found in %s", filename)


class Validate(Runner):
    """
    The `validate` command for checking that .po files have no errors.
    """

    def add_args(self):
        self.parser.description = "Automatically finds translation errors in all edx-platform *.po files, "\
            "for all languages, unless one or more language(s) is specified to check."

        self.parser.add_argument(
            '-l', '--language',
            type=str,
            nargs='*',
            help="Specify one or more specific language code(s) to check (eg 'ko_KR')."
        )

        self.parser.add_argument(
            '-e', '--empty',
            action='store_true',
            help="Includes empty translation strings in .prob files."
        )

    def run(self, args):
        """
        Main entry point for script
        """
        if args.verbose:
            log_level = logging.INFO
        else:
            log_level = logging.WARNING
        logging.basicConfig(stream=sys.stdout, level=log_level)

        languages = args.language or []
        locale_dir = self.configuration.locale_dir
        if not languages:
            # validate all languages
            validate_po_files(self.configuration, locale_dir, args.empty)
        else:
            # languages will be a list of language codes; test each language.
            for language in languages:
                root_dir = self.configuration.locale_dir / language
                # Assert that a directory for this language code exists on the system
                if not root_dir.isdir():
                    log.error(" %s is not a valid directory.\nSkipping language '%s'", root_dir, language)
                    continue
                # If we found the language code's directory, validate the files.
                validate_po_files(self.configuration, locale_dir, root_dir=root_dir, report_empty=args.empty)

main = Validate()  # pylint: disable=invalid-name

if __name__ == '__main__':
    print("Validating languages...")
    main()
    print("Finished validating languages")
